import { solveTable } from "../solver.js";
import { Table } from "../../lib/table.js";
import { FormulaError } from "../evaluator.js";
import dayjs from "../../_virtual/dayjs.min.js";
import { FULLDATE_FORMAT_UTC } from "../../constants.js";
const gt = (left, right) => {
  if (typeof left === "string" || typeof right === "string") {
    return ensureString(left) > ensureString(right);
  }
  try {
    return ensureNumber(left) > ensureNumber(right);
  } catch {
    return false;
  }
};
const gte = (left, right) => {
  if (typeof left === "string" || typeof right === "string") {
    return ensureString(left) >= ensureString(right);
  }
  try {
    return ensureNumber(left) >= ensureNumber(right);
  } catch {
    return false;
  }
};
const lt = (left, right) => {
  return !gte(left, right);
};
const lte = (left, right) => {
  return !gt(left, right);
};
const eq = (left, right) => {
  return ensureString(left) === ensureString(right);
};
const ne = (left, right) => {
  return !eq(left, right);
};
const ensureNumber = (value, alternative) => {
  if (!value) {
    return 0;
  }
  if (value instanceof Table) {
    const v = stripTable(value, 0, 0);
    return ensureNumber(v);
  }
  if (value instanceof Date) {
    return value.getTime();
  }
  const num = parseFloat(value);
  if (isNaN(num)) {
    throw new FormulaError("#VALUE!", `${value} cannot be converted to a number`);
  }
  return num;
};
const ensureString = (value) => {
  if (value === 0) {
    return "0";
  }
  if (!value) {
    return "";
  }
  if (value instanceof Table) {
    const v = stripTable(value, 0, 0);
    return ensureString(v);
  }
  switch (value.constructor.name) {
    case "Date": {
      const d = value;
      return dayjs(d).format(FULLDATE_FORMAT_UTC);
    }
    default:
      return String(value);
  }
};
const ensureBoolean = (value, alternative) => {
  if (typeof value === "undefined" && typeof alternative !== "undefined") {
    return alternative;
  }
  if (value === null) {
    return false;
  }
  if (value instanceof Table) {
    const v = stripTable(value, 0, 0);
    return ensureBoolean(v, alternative);
  }
  if (typeof value === "string" || value instanceof String) {
    const bool = { true: true, false: false }[value.toLowerCase()];
    if (bool == null) {
      throw new FormulaError("#VALUE!", `text '${value}' cannot be converted to a boolean`);
    }
    return bool;
  }
  return Boolean(value);
};
const stripTable = (value, y = 0, x = 0) => {
  if (value instanceof Table) {
    return solveTable({ table: value })[y][x];
  }
  return value;
};
const CONDITION_REGEX = /^(<=|>=|<>|>|<|=)?(.*)$/;
const check = (value, condition) => {
  const m = condition.match(CONDITION_REGEX);
  const [, expr = "", target = ""] = m || [];
  let comparison = target;
  if (expr === ">" || expr === "<" || expr === ">=" || expr === "<=") {
    if (typeof value === "number") {
      comparison = parseFloat(target);
    }
    switch (expr) {
      case ">":
        return gt(value, comparison);
      case ">=":
        return gte(value, comparison);
      case "<":
        return lt(value, comparison);
      case "<=":
        return lte(value, comparison);
    }
  }
  const equals = expr === "" || expr === "=";
  if (target === "") {
    return (value == null || value === "") === equals;
  }
  if (typeof value === "string" || value instanceof String) {
    const replaced = target.replace(/~\*/g, "(\\*)").replace(/~\?/g, "(\\?)").replace(/\*/g, "(.*)").replace(/\?/g, "(.)");
    const regex = RegExp(`^${replaced}$`, "i");
    return regex.test(value) === equals;
  }
  return eq(value, comparison) === equals;
};
export {
  check,
  ensureBoolean,
  ensureNumber,
  ensureString,
  eq,
  gt,
  gte,
  lt,
  lte,
  ne,
  stripTable
};
//# sourceMappingURL=__utils.js.map
