import { a2p, c2x, x2c, p2a } from "./converters.js";
const superposeArea = (srcArea, dstArea) => {
  const { height: srcHeight, width: srcWidth } = areaShape(srcArea);
  const { height: dstHeight, width: dstWidth } = areaShape(dstArea);
  return {
    height: srcHeight > dstHeight ? srcHeight : dstHeight,
    width: srcWidth > dstWidth ? srcWidth : dstWidth
  };
};
const concatAreas = (area1, area2) => {
  const result = { ...area1 };
  if (area2.left < area1.left) {
    result.left = area2.left;
  }
  if (area2.right > area1.right) {
    result.right = area2.right;
  }
  if (area2.top < area1.top) {
    result.top = area2.top;
  }
  if (area2.bottom > area1.bottom) {
    result.bottom = area2.bottom;
  }
  return result;
};
const zoneToArea = (zone) => {
  if (zone.endY === -1 || zone.endX === -1) {
    return { top: -1, left: -1, bottom: -1, right: -1 };
  }
  const [top, bottom] = zone.startY < zone.endY ? [zone.startY, zone.endY] : [zone.endY, zone.startY];
  const [left, right] = zone.startX < zone.endX ? [zone.startX, zone.endX] : [zone.endX, zone.startX];
  return { top, left, bottom, right };
};
const areaToZone = (area) => {
  return {
    startY: area.top,
    startX: area.left,
    endY: area.bottom,
    endX: area.right
  };
};
const areaToRange = (area) => {
  const { top, left, bottom, right } = area;
  return `${p2a({ y: top, x: left })}:${p2a({
    y: bottom,
    x: right
  })}`;
};
const between = (range2, index) => {
  if (range2.start === -1 || range2.end === -1) {
    return false;
  }
  return range2.start <= index && index <= range2.end || range2.end <= index && index <= range2.start;
};
const among = (area, point) => {
  if (area.top === -1 || area.left === -1 || area.bottom === -1 || area.right === -1) {
    return false;
  }
  const { y, x } = point;
  const { top, left, bottom, right } = area;
  return top <= y && y <= bottom && left <= x && x <= right;
};
const zoneShape = ({ base = 0, ...zone }) => {
  return {
    height: base + Math.abs(zone.startY - zone.endY),
    width: base + Math.abs(zone.startX - zone.endX)
  };
};
const areaShape = ({ base = 0, ...area }) => {
  return {
    height: base + Math.abs(area.top - area.bottom),
    width: base + Math.abs(area.left - area.right)
  };
};
const matrixShape = ({ base = 0, matrix }) => {
  const h = matrix.length;
  if (h === 0) {
    return { height: 0, width: 0 };
  }
  return { height: base + h, width: base + matrix[0].length };
};
const makeSequence = (start, stop, step = 1) => {
  return Array.from({ length: (stop - start - 1) / step + 1 }, (_, i) => start + i * step);
};
const oa2aa = (oa, fields) => {
  const aa = [];
  oa.forEach((o) => {
    const a = [];
    fields.forEach((field) => {
      a.push(o[field]);
    });
    aa.push(a);
  });
  return aa;
};
const aa2oa = (aa, fields) => {
  const oa = [];
  aa.forEach((a) => {
    const o = {};
    a.forEach((v, i) => {
      if (i >= fields.length) {
        return;
      }
      const field = fields[i];
      o[field] = v;
    });
    oa.push(o);
  });
  return oa;
};
const putMatrix = (dst, src, dstArea, filter = () => true) => {
  const lostRows = {};
  const { top, left, bottom, right } = dstArea;
  const { height: dstNumRows, width: dstNumCols } = matrixShape({
    matrix: dst,
    base: 1
  });
  for (let y = top; y <= bottom; y++) {
    const lostRow = [];
    for (let x = left; x <= right; x++) {
      const value = src[y - top][x - left];
      if (y < dstNumRows - 1 && x < dstNumCols - 1) {
        if (filter(value, dst[y][x])) {
          dst[y][x] = value;
        }
        continue;
      }
      if (lostRow.length === 0) {
        lostRows[p2a({ y, x })] = [lostRow];
      }
      lostRow.push(value);
    }
  }
  return lostRows;
};
const createMatrix = (numRows, numCols, fill) => {
  return [...Array(numRows)].map(() => Array(numCols).fill(fill));
};
const constructInitialCellsOrigin = ({
  cells = {},
  ensured = {},
  matrix = [],
  flattenAs = "value"
}) => {
  return constructInitialCells({
    cells,
    ensured,
    matrices: { A1: matrix },
    flattenAs
  });
};
const constructInitialCells = ({
  cells = {},
  ensured = {},
  matrices = {},
  flattenAs = "value"
} = {}) => {
  upsert({ cells, flattenAs, matrices });
  const { numRows, numCols } = Object.assign({ numRows: 1, numCols: 1 }, ensured);
  const rightBottom = p2a({ y: numRows, x: numCols });
  if (cells[rightBottom] == null) {
    cells[rightBottom] = {};
  }
  return cells;
};
const upsert = ({
  cells = {},
  matrices = {},
  flattenAs
}) => {
  Object.keys(matrices).forEach((baseAddress) => {
    const matrix = matrices[baseAddress];
    const { y: baseY, x: baseX } = a2p(baseAddress);
    matrix.forEach((row, y) => {
      row.forEach((e, x) => {
        const id = p2a({ y: baseY + y, x: baseX + x });
        if (flattenAs) {
          const cell = cells[id];
          cells[id] = { [flattenAs]: e, ...cell };
        } else {
          cells[id] = e;
        }
      });
    });
  });
  return cells;
};
const getMaxSizesFromCells = (cells = {}) => {
  let [lastY, lastX] = [0, 0];
  Object.keys(cells).map((address) => {
    const { y, x } = a2p(address);
    if (lastY < y) {
      lastY = y;
    }
    if (lastX < x) {
      lastX = x;
    }
  });
  return { numRows: lastY, numCols: lastX };
};
const range = (start, end) => {
  const list = [];
  for (let i = start; i <= end; i++) {
    list.push(i);
  }
  return list;
};
const complementSelectingArea = (selectingArea, choosing) => {
  if (selectingArea.left === -1) {
    selectingArea = {
      left: choosing.x,
      top: choosing.y,
      right: choosing.x,
      bottom: choosing.y
    };
  }
  return selectingArea;
};
const expandRange = (range2) => {
  if (range2.indexOf(":") === -1) {
    return [range2];
  }
  const result = [];
  const isRowRange = /^\d+\:\d+$/.test(range2);
  if (isRowRange) {
    const [startRow2, endRow2] = range2.split(":").map(Number);
    for (let row = startRow2; row <= endRow2; row++) {
      result.push(`${row}`);
    }
    return result;
  }
  const match = range2.match(/^([A-Z]*)(\d+)?\:([A-Z]*)(\d+)?$/);
  if (!match) {
    console.error("Invalid range format", range2);
    return [range2];
  }
  const [, startCol, startRow, endCol, endRow] = match;
  const startColIndex = startCol ? c2x(startCol) : 1;
  const endColIndex = endCol ? c2x(endCol) : 1;
  for (let col = startColIndex; col <= endColIndex; col++) {
    const currentColumn = startCol || endCol ? x2c(col) : "";
    if (startRow && endRow) {
      for (let row = Number(startRow); row <= Number(endRow); row++) {
        result.push(`${currentColumn}${row}`);
      }
    } else if (!startRow && !endRow) {
      result.push(currentColumn);
    }
  }
  return result;
};
export {
  aa2oa,
  among,
  areaShape,
  areaToRange,
  areaToZone,
  between,
  complementSelectingArea,
  concatAreas,
  constructInitialCells,
  constructInitialCellsOrigin,
  createMatrix,
  expandRange,
  getMaxSizesFromCells,
  makeSequence,
  matrixShape,
  oa2aa,
  putMatrix,
  range,
  superposeArea,
  upsert,
  zoneShape,
  zoneToArea
};
//# sourceMappingURL=structs.js.map
