import { TimeDelta } from "../lib/time.js";
import dayjs from "../_virtual/dayjs.min.js";
import utc from "../_virtual/utc.js";
import timezone from "../_virtual/timezone.js";
dayjs.extend(timezone);
dayjs.extend(utc);
const BOOLS = { true: true, false: false };
const NUMS = /* @__PURE__ */ new Set(["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]);
const NUMS_Z = /* @__PURE__ */ new Set([...NUMS, "Z", "z"]);
const JFMASOND = /* @__PURE__ */ new Set(["J", "F", "M", "A", "S", "O", "N", "D", ...NUMS]);
const NBRYNLGPTVC = /* @__PURE__ */ new Set(["N", "B", "R", "Y", "N", "L", "G", "P", "T", "V", "C", ...NUMS_Z]);
class Parser {
  constructor(props) {
    this.functions = [
      this.number.bind(this),
      this.timedelta.bind(this),
      this.date.bind(this),
      this.bool.bind(this)
    ];
    this.applyMixins(props == null ? void 0 : props.mixins);
    if (props == null) {
      return;
    }
    const { condition, complement } = props;
    this.condition = condition;
    this.complement = complement;
  }
  applyMixins(mixins) {
    if (mixins == null) {
      return;
    }
    for (const mixin of mixins) {
      for (const key in mixin) {
        this[key] = mixin[key];
      }
    }
  }
  call(value, cell) {
    try {
      const parsed = this.parse(value, cell);
      return this.callback(parsed, cell);
    } catch (e) {
      return this.callback(e, cell);
    }
  }
  callback(parsed, cell) {
    return { ...cell, value: parsed };
  }
  parse(value, cell) {
    if (this.condition && !this.condition(value)) {
      const result = this.complement ? this.complement(value) : value;
      return result;
    }
    if (value[0] === "'") {
      return value;
    }
    for (let i = 0; i < this.functions.length; i++) {
      const result = this.functions[i](value, cell);
      if (result != null) {
        return result;
      }
    }
    if (value === "") {
      return null;
    }
    return value;
  }
  bool(value, cell) {
    return BOOLS[value.toLowerCase()];
  }
  number(value, cell) {
    const m = value.match(/^-?[\d.]+$/);
    if (m != null && value.match(/\.$/) == null && (value.match(/\./g) || []).length <= 1) {
      return parseFloat(value);
    }
  }
  timedelta(value, cell) {
    if (value.length < 4 || isNaN(value[value.length - 1])) {
      return;
    }
    return TimeDelta.parse(value);
  }
  date(value, cell) {
    const first = value[0];
    if (first == null || !JFMASOND.has(first.toUpperCase())) {
      return;
    }
    if (!NBRYNLGPTVC.has(value[value.length - 1].toUpperCase())) {
      return;
    }
    if (value.match(/[=*&#@!?[\]{}"'()|%\\<>~+\r\n]/)) {
      return;
    }
    let timeZone = "UTC";
    try {
      timeZone = Intl.DateTimeFormat().resolvedOptions().timeZone;
    } catch (e) {
    }
    try {
      const day = dayjs.tz(value, timeZone);
      return day.toDate();
    } catch (e) {
    }
  }
}
const defaultParser = new Parser();
export {
  Parser,
  defaultParser
};
//# sourceMappingURL=core.js.map
