import dayjs from "../_virtual/dayjs.min.js";
import { Table } from "../lib/table.js";
import { solveFormula } from "../formula/solver.js";
import { FormulaError } from "../formula/evaluator.js";
import { p2a } from "../lib/converters.js";
import { TimeDelta } from "../lib/time.js";
class Renderer {
  constructor(props) {
    this.datetimeFormat = "YYYY-MM-DD HH:mm:ss";
    this.dateFormat = "YYYY-MM-DD";
    this.timeDeltaFormat = "HH:mm:ss";
    this.applyMixins(props == null ? void 0 : props.mixins);
    if (props == null) {
      return;
    }
    const { condition, complement } = props;
    this.condition = condition;
    this.complement = complement;
  }
  applyMixins(mixins) {
    if (mixins == null) {
      return;
    }
    for (const mixin of mixins) {
      for (const key in mixin) {
        this[key] = mixin[key];
      }
    }
  }
  call(table, point, writer) {
    const address = p2a(point);
    const cache = table.getSolvedCache(address);
    const cell = table.getByPoint(point);
    let value = cache || (cell == null ? void 0 : cell.value);
    if (typeof value === "string" && !(cell == null ? void 0 : cell.disableFormula)) {
      if (value[0] === "'") {
        value = value.substring(1);
      } else if (value[0] === "=") {
        value = solveFormula({ value, table, raise: true, origin: point });
      }
    }
    return this.render(value, table, writer, point);
  }
  render(value, table, writer, position) {
    var _a;
    if (this.condition && !this.condition(value)) {
      return this.complement ? this.complement(value) : this.stringify({ value });
    }
    switch (typeof value) {
      case "object":
        if (value instanceof Date) {
          return this.date(value, writer, position);
        }
        if (TimeDelta.is(value)) {
          return this.timedelta(TimeDelta.ensure(value), writer, position);
        }
        if (value == null) {
          return this.null(value, writer, position);
        }
        if (value instanceof Table) {
          return this.render((_a = value.getByPoint({ y: value.top, x: value.left })) == null ? void 0 : _a.value, table, writer, position);
        }
        if (Array.isArray(value)) {
          return this.array(value, writer, position);
        }
        if (value instanceof FormulaError) {
          throw value;
        }
        return this.object(value, writer, position);
      case "string":
        return this.string(value, table, writer, position);
      case "number":
        return this.number(value, writer, position);
      case "boolean":
        return this.bool(value, writer, position);
      case "undefined":
        return this.undefined(value, writer, position);
      case "function":
        return value();
    }
    return "";
  }
  stringify(cell, position) {
    const { value } = cell;
    if (value instanceof Date) {
      return this.date(value);
    }
    if (TimeDelta.is(value)) {
      return this.timedelta(TimeDelta.ensure(value));
    }
    if (value == null) {
      return "";
    }
    return value.toString();
  }
  string(value, table, writer, position) {
    return value;
  }
  bool(value, writer, position) {
    return value ? "TRUE" : "FALSE";
  }
  number(value, writer, position) {
    if (isNaN(value)) {
      return "NaN";
    }
    return value;
  }
  date(value, writer, position) {
    if (value.getHours() + value.getMinutes() + value.getSeconds() === 0) {
      return dayjs(value).format(this.dateFormat);
    }
    return dayjs(value).format(this.datetimeFormat);
  }
  timedelta(value, writer, position) {
    return value.stringify(this.timeDeltaFormat);
  }
  array(value, writer, position) {
    return value.map((v) => this.stringify({ value: v })).join(",");
  }
  object(value, writer, position) {
    return JSON.stringify(value);
  }
  null(value, writer, position) {
    return "";
  }
  undefined(value, writer, position) {
    return "";
  }
}
const defaultRenderer = new Renderer();
export {
  Renderer,
  defaultRenderer
};
//# sourceMappingURL=core.js.map
